package envoy.client.ui;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.swing.JOptionPane;

import envoy.client.data.*;
import envoy.client.net.Client;
import envoy.client.net.WriteProxy;
import envoy.data.Message;
import envoy.data.User;
import envoy.data.User.UserStatus;
import envoy.exception.EnvoyException;
import envoy.util.EnvoyLog;
import javafx.application.Application;
import javafx.fxml.FXMLLoader;
import javafx.scene.Scene;
import javafx.scene.image.Image;
import javafx.scene.layout.GridPane;
import javafx.stage.Stage;

/**
 * Project: <strong>envoy-client</strong><br>
 * File: <strong>Startup.java</strong><br>
 * Created: <strong>26.03.2020</strong><br>
 *
 * @author Kai S. K. Engelbart
 * @since Envoy Client v0.1-beta
 */
public final class Startup extends Application {

	private LocalDB			localDB;
	private Client			client;
	private WriteProxy		writeProxy;
	private Cache<Message>	cache;

	private static final ClientConfig	config	= ClientConfig.getInstance();
	private static final Logger			logger	= EnvoyLog.getLogger(Startup.class);

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void init() throws Exception {
		try {
			// Load the configuration from client.properties first
			Properties properties = new Properties();
			properties.load(Startup.class.getClassLoader().getResourceAsStream("client.properties"));
			config.load(properties);

			// Override configuration values with command line arguments
			String[] args = getParameters().getRaw().toArray(new String[0]);
			if (args.length > 0) config.load(args);

			// Check if all mandatory configuration values have been initialized
			if (!config.isInitialized()) throw new EnvoyException("Configuration is not fully initialized");
		} catch (Exception e) {
			JOptionPane.showMessageDialog(null, "Error loading configuration values:\n" + e, "Configuration error", JOptionPane.ERROR_MESSAGE);
			e.printStackTrace();
			System.exit(1);
		}

		// Setup logger for the envoy package
		EnvoyLog.initialize(config);
		EnvoyLog.attach("envoy");
		EnvoyLog.setFileLevelBarrier(config.getFileLevelBarrier());
		EnvoyLog.setConsoleLevelBarrier(config.getConsoleLevelBarrier());

		// Initialize the local database
		if (config.isIgnoreLocalDB()) {
			localDB = new TransientLocalDB();
			JOptionPane.showMessageDialog(null,
					"Ignoring local database.\nMessages will not be saved!",
					"Local database warning",
					JOptionPane.WARNING_MESSAGE);
		} else try {
			localDB = new PersistentLocalDB(new File(config.getHomeDirectory(), config.getLocalDB().getPath()));
		} catch (IOException e3) {
			logger.log(Level.SEVERE, "Could not initialize local database", e3);
			JOptionPane.showMessageDialog(null, "Could not initialize local database!\n" + e3, "Local database error", JOptionPane.ERROR_MESSAGE);
			System.exit(1);
			return;
		}

		// Initialize client and unread message cache
		client	= new Client();
		cache	= new Cache<>();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void start(Stage stage) throws Exception {

		// Try to connect to the server
		// new LoginDialog(client, localDB, cache);
		new LoginDialog().showAndWait();
		System.exit(0);

		// Set client user in local database
		localDB.setUser(client.getSender());

		// Initialize chats in local database
		try {
			localDB.initializeUserStorage();
			localDB.loadUserData();
		} catch (FileNotFoundException e) {
			// The local database file has not yet been created, probably first login
		} catch (Exception e) {
			e.printStackTrace();
			JOptionPane.showMessageDialog(null,
					"Error while loading local database: " + e + "\nChats will not be stored locally.",
					"Local DB error",
					JOptionPane.WARNING_MESSAGE);
		}

		// Initialize write proxy
		writeProxy = client.createWriteProxy(localDB);

		if (client.isOnline()) {

			// Save all users to the local database and flush cache
			localDB.setUsers(client.getUsers());
			writeProxy.flushCache();
		} else
			// Set all contacts to offline mode
			localDB.getUsers()
				.values()
				.stream()
				.filter(u -> u instanceof User && u != localDB.getUser())
				.map(User.class::cast)
				.forEach(u -> u.setStatus(UserStatus.OFFLINE));

		// Prepare stage and load ChatScene
		var	loader		= new FXMLLoader(getClass().getResource("ChatScene.fxml"));
		var	anchorPane	= loader.<GridPane>load();
		var	chatScene	= new Scene(anchorPane);
		stage.setTitle("Envoy");
		stage.getIcons().add(new Image(getClass().getResourceAsStream("/icons/envoy_logo.png")));
		stage.setScene(chatScene);

		loader.<ChatSceneController>getController().initializeData(localDB, client, writeProxy);
		stage.show();

		// Relay unread messages from cache
		if (cache != null && client.isOnline()) cache.relay();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void stop() throws Exception {
		try {

			// Save Settings and PersistentLocalDB on shutdown
			logger.info("Closing connection...");
			client.close();

			logger.info("Saving local database and settings...");
			localDB.save();
			Settings.getInstance().save();
		} catch (Exception e) {
			logger.log(Level.SEVERE, "Unable to save local files", e);
		}
	}

	@SuppressWarnings("javadoc")
	public static void main(String[] args) { launch(args); }
}
